/*

*/

import java.util.concurrent.Future;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.TimeUnit;

/**
   Implement a version of Future that we can use in our
   MyExecutorService thread pool.

   This "future" is more like a "promise" or what Java
   calls a CompletableFuture.
*/
public class MyFuture<V> implements Future<V>
{
   private final ReentrantLock lock = new ReentrantLock();
   private final Condition completed = lock.newCondition();
   private V result = null;

   public boolean isDone()
   {
      return null != result;
   }


   public V get() throws InterruptedException
   {
      lock.lock();
      try
      {
         while (null == result)
         {
            completed.await();
         }
         return result;
      }
      finally
      {
         lock.unlock();
      }
   }


   public V get(long timeout, TimeUnit unit) throws InterruptedException
   {
      lock.lock();
      try
      {
         if (null == result)
         {
            completed.await(timeout, unit);
         }
         return result;
      }
      finally
      {
         lock.unlock();
      }
   }


   public boolean complete(V value)
   {
      lock.lock();
      try
      {
         if (null == result)
         {
            result = value;
            completed.signalAll();
            return true;
         }
         else
         {
            return false;
         }
      }
      finally
      {
         lock.unlock();
      }
   }

   public boolean cancel(boolean mayInterruptIfRunning) { return false; }
   public boolean isCancelled() { return false; }
}
