
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph using four threads
   and "message passing". This will not deadlock.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4
*/
public class TaskGraph_v2
{
   public static void main(String[] args)
   {
      final Task<Integer,Integer> task1 = new Task<>(1, 1, 2);
      final Task<Integer,Integer> task2 = new Task<>(2, 2, 1);
      final Task<Integer,Integer> task3 = new Task<>(3, 3, 3);
      final Task<Integer,Integer> task4 = new Task<>(4, 4, 2);

      final CompletableFuture<Integer> future1 = new CompletableFuture<>(); // empty future
      final CompletableFuture<Integer> future2 = new CompletableFuture<>(); // empty future
      final CompletableFuture<Integer> future3 = new CompletableFuture<>(); // empty future

      final Thread t1 = new Thread(
                          ()->{ task1.run();
                                future1.complete(0); }   // Send a message to tasks 2 and 3.
                                  );

      final Thread t2 = new Thread(
                          ()->{try{ future1.get();       // Receive a message from task1.
                                    task2.run();
                                    future2.complete(0); // Send a message to task4.
                                  }catch(InterruptedException
                                        |ExecutionException e){} }
                                  );

      final Thread t3 = new Thread(
                          ()->{try{ future1.get();       // Receive a message from task1.
                                    task3.run();
                                    future3.complete(0); // Send a message to task4.
                                  }catch(InterruptedException
                                        |ExecutionException e){} }
                                  );

      final Thread t4 = new Thread(
                          ()->{try{ future2.get(); // Receive a message from task2.
                                    future3.get(); // Receive a message from task3.
                                    task4.run();
                                  }catch(InterruptedException
                                        |ExecutionException e){} }
                                  );
      t4.start();
      t3.start();
      t2.start();
      t1.start();
   }
}
