
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph on a thread pool.
   This implementation uses "message passing"
   (on a thread pool). This can deadlock.

   Compare this to TaskGraph_v2.java, which uses
   the exact same message passing structure, but
   TaskGraph_v2 cannot deadlock because it isn't
   on a thread pool.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4
*/
public class TaskGraph_v4
{
   public static void main(String[] args)
   {
      final int threads = 3;
      final MyThreadPool_v1 executor = new MyThreadPool_v1(threads);

      final Task<Integer,Integer> task1 = new Task<>(1, 1, 2);
      final Task<Integer,Integer> task2 = new Task<>(2, 2, 1);
      final Task<Integer,Integer> task3 = new Task<>(3, 3, 3);
      final Task<Integer,Integer> task4 = new Task<>(4, 4, 2);

      final CompletableFuture<Integer> future1 = new CompletableFuture<>(); // empty future
      final CompletableFuture<Integer> future2 = new CompletableFuture<>(); // empty future
      final CompletableFuture<Integer> future3 = new CompletableFuture<>(); // empty future


      final Runnable r1 = ()->{ task1.run();
                                future1.complete(0);     // Send a message to tasks 2 and 3.
                              };

      final Runnable r2 = ()->{try{ future1.get();       // Receive a message from task1.
                                    task2.run();
                                    future2.complete(0); // Send a message to task4.
                                  }catch(InterruptedException
                                        |ExecutionException e){}
                              };

      final Runnable r3 = ()->{try{ future1.get();       // Receive a message from task1.
                                    task3.run();
                                    future3.complete(0); // Send a message to taks4.
                                  }catch(InterruptedException
                                        |ExecutionException e){}
                              };

      final Runnable r4 = ()->{try{ future2.get(); // Receive a message from taks2.
                                    future3.get(); // Receive a message from task2.
                                    task4.run();
                                  }catch(InterruptedException
                                        |ExecutionException e){}
                              };


      executor.execute( r4 );
      executor.execute( r3 );
      executor.execute( r2 );
      executor.execute( r1 );

      executor.shutdown();
   }
}
