
import java.util.concurrent.Future;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph on an executor service
   style thread pool using Callable tasks and Futures.

   This cannot deadlock?

   Compare this with version TaskGraph_v4.java. They
   have the exact same message passing structure, and
   they both run on a thread pool, but TaskGraph_v4
   can deadlock, and this version cannot. The use
   of final futures somehow forces a deadlock free
   ordering of the tasks in the thread pool's task
   queue.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4
*/
public class TaskGraph_v6
{
   public static void main(String[] args)
   {
      final int threads = 2;
      final var executor = new MyExecutorService<Integer>(threads);
    //final ExecutorService executor = Executors.newFixedThreadPool(threads);

      final TaskC<Integer,Integer> task1 = new TaskC<>(1, 1, 2);
      final TaskC<Integer,Integer> task2 = new TaskC<>(2, 2, 1);
      final TaskC<Integer,Integer> task3 = new TaskC<>(3, 3, 3);
      final TaskC<Integer,Integer> task4 = new TaskC<>(4, 4, 2);

      final Future<Integer> future1 = executor.submit(task1); // Send a message to tasks 2 and 3.

      final Future<Integer> future2 = executor.submit(
                          ()->{ future1.get();              // Receive a message from task1.
                                return task2.call(); });    // Send a message to task4.

      final Future<Integer> future3 = executor.submit(
                          ()->{ future1.get();              // Receive a message from taks1.
                                return task3.call(); });    // Send a message to task4.

      final Future<Integer> future4 = executor.submit(
                          ()->{ future2.get();              // Receive a message from task2.
                                future3.get();              // Receive a message from task3.
                                return task4.call(); });    // Send a message to main thread.

      try{ future4.get(); }
      catch(InterruptedException|ExecutionException e){}

      executor.shutdown();
   }
}
