
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Future;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph on an executor service
   thread pool using Callable tasks, Futures, a
   CompletableFuture, and fewer tasks than TaskGraph_v6.
   Now we are back to the deadlockk problem!

   Compare this to TaskGraph_v4_v2.java, which uses
   the exact same message passing structure.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4
*/
public class TaskGraph_v6_v2
{
   public static void main(String[] args)
   {
      final int threads = 2;  // try 1 thread
      final ExecutorService executor = Executors.newFixedThreadPool(threads);

      final TaskC<Integer,Integer> task1 = new TaskC<>(1, 1, 2);
      final TaskC<Integer,Integer> task2 = new TaskC<>(2, 2, 1);
      final TaskC<Integer,Integer> task3 = new TaskC<>(3, 3, 3);
      final TaskC<Integer,Integer> task4 = new TaskC<>(4, 4, 2);

      final CompletableFuture<Integer> future = new CompletableFuture<>(); // empty future

      final Future<Integer> future1 = executor.submit(
                          ()->{ future.get(); // Receive a message.
                                return task3.call(); });

      final Future<Integer> future2 = executor.submit(
                          ()->{ future.complete(task1.call()); // Send a message.
                                task2.call();
                                future1.get();
                                return task4.call(); });

      try{ future2.get(); }
      catch(InterruptedException|ExecutionException e){}

      executor.shutdown();
   }
}
