
import java.util.concurrent.Future;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Callable;

/**
   Implement this task graph on an executor service
   style thread pool using Callable tasks and Futures.

   This version is the same as the previous version.
   This version just separates the declarations of the
   tasks from the declarations of the futures. This kind
   of highlights the way their order in the thread pool's
   task queue is forced on us.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4

   It is very hard to cause a deadlock
   when using this style of programming.
*/
public class TaskGraph_v6a
{
   public static void main(String[] args)
   {
      final int threads = 2;
      final var executor = new MyExecutorService<Integer>(threads);
    //final ExecutorService executor = Executors.newFixedThreadPool(threads);

      final TaskC<Integer,Integer> task1 = new TaskC<>(1, 1, 2);
      final TaskC<Integer,Integer> task2 = new TaskC<>(2, 2, 1);
      final TaskC<Integer,Integer> task3 = new TaskC<>(3, 3, 3);
      final TaskC<Integer,Integer> task4 = new TaskC<>(4, 4, 2);

      final Future<Integer> future1 = executor.submit(task1);

      final Callable<Integer> task5 =
                          ()->{ future1.get();
                                return task2.call(); };

      final Callable<Integer> task6 =
                          ()->{ future1.get();
                                return task3.call(); };

      final Future<Integer> future2 = executor.submit(task5);
      final Future<Integer> future3 = executor.submit(task6);

      final Callable<Integer> task7 =
                          ()->{ future2.get();
                                future3.get();
                                return task4.call(); };

      final Future<Integer> future4 = executor.submit(task7);

      try{ future4.get(); }
      catch(InterruptedException|ExecutionException e){}

      executor.shutdown();
   }
}
