
import java.util.concurrent.Future;
import java.util.concurrent.Executors;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.ExecutionException;

/**
   Implement this task graph on an executor service
   style thread pool using Callable tasks and Futures.

              task1
             /     \
            /       \
       task2         task3
            \       /
             \     /
              task4

   It is very hard to cause a deadlock
   when using this style of programming.

   If we break the rule that the executir service
   must create the Future object, then we can cause
   a deadlock.
*/
public class TaskGraph_v6c
{
   public static void main(String[] args)
   {
      final int threads = 2; // try 2 threads
      final var executor = new MyExecutorService<Integer>(threads);


      final TaskC<Integer,Integer> task1 = new TaskC<>(1, 1, 2);
      final TaskC<Integer,Integer> task2 = new TaskC<>(2, 2, 1);
      final TaskC<Integer,Integer> task3 = new TaskC<>(3, 3, 3);
      final TaskC<Integer,Integer> task4 = new TaskC<>(4, 4, 2);


      final MyFuture<Integer> future2 = new MyFuture<>(); // empty future
      final MyFuture<Integer> future3 = new MyFuture<>(); // empty future

      final Future<Integer> future1 = executor.submit(
                          ()->{ final Integer i = task1.call();
                                executor.submit(task2, future2); // A new (bad) version of submit().
                                executor.submit(task3, future3); // A new (bad) version of submit().
                                return i; });

      final Future<Integer> future4 = executor.submit( // This task is in the task queue ahead of task2.
                          ()->{ future2.get();
                                future3.get();
                                return task4.call(); });

      try{ future4.get(); }
      catch(InterruptedException|ExecutionException e){}

      executor.shutdown();
   }
}
